package de.fraunhofer.iem.dataprovider.taskManager

import de.fraunhofer.iem.dataprovider.configuration.Config
import de.fraunhofer.iem.dataprovider.configuration.OpenCodeGitlabApiConfiguration
import de.fraunhofer.iem.dataprovider.dependency.service.DependencyService
import de.fraunhofer.iem.dataprovider.kpi.service.KPIService
import de.fraunhofer.iem.dataprovider.logger.getLogger
import de.fraunhofer.iem.dataprovider.repository.service.RepositoryService
import de.fraunhofer.iem.dataprovider.taskManager.events.*
import de.fraunhofer.iem.dataprovider.taskManager.tasks.dataQuery.CloneGitTask
import de.fraunhofer.iem.dataprovider.taskManager.tasks.dataQuery.GetGitlabProjectTask
import de.fraunhofer.iem.dataprovider.taskManager.tasks.dataQuery.GetRepositoryDetailsTask
import de.fraunhofer.iem.dataprovider.taskManager.tasks.kpiCalculation.MetricsTask
import de.fraunhofer.iem.dataprovider.taskManager.tasks.tools.occmd.OccmdTask
import de.fraunhofer.iem.dataprovider.taskManager.tasks.tools.ort.OrtAdvisorTask
import de.fraunhofer.iem.dataprovider.taskManager.tasks.tools.ort.OrtAnalyzerTask
import de.fraunhofer.iem.dataprovider.tool.service.ToolService
import de.fraunhofer.iem.dataprovider.toolRun.service.ToolRunService
import jakarta.annotation.PreDestroy
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.ObsoleteCoroutinesApi
import kotlinx.coroutines.channels.SendChannel
import kotlinx.coroutines.channels.actor
import org.springframework.stereotype.Component


/**
 * The Task Manager takes tasks and distributes them to
 * underlying workers. Internally it uses a channel
 * to manage incoming tasks.
 */
@Component
class TaskManager(
    private val config: Config,
    private val openCodeGitlabConfiguration: OpenCodeGitlabApiConfiguration,
    private val toolRunService: ToolRunService,
    private val repositoryService: RepositoryService,
    private val kpiService: KPIService,
    private val dependencyService: DependencyService,
    private val toolService: ToolService
) {

    // The used default dispatcher is ok for CPU-bound workloads. However,
    // if they block for a long time it's better to use a custom thread
    // pool solution.
    private val worker = Worker("Worker")

    // Should be used for long-running tasks, which DON'T use CPU power and
    // are non-blocking
    private val ioWorker = Worker("IO-Worker", CoroutineScope(Dispatchers.IO))
    // Should be used for long-running CPU heavy tasks, which potentially block.
    // private val customWorker = getCustomThreadpoolWorker()

    private val mainScope = CoroutineScope(Dispatchers.Default)
    private val actor: SendChannel<Event> = mainScope.taskActor()

    suspend fun addEvent(event: Event) {
        actor.send(event)
    }

    @OptIn(ObsoleteCoroutinesApi::class)  // This API is meant to be improved since 2017... so this should be good to use.
    // IntelliJ might tell u that it is ok to remove <ITask> from the actor declaration.
    // IT IS NOT! this will lead to very strange gradle compile errors!
    private fun CoroutineScope.taskActor() = actor<Event> {
        with(MyActor()) {
            for (task in channel) {
                onReceive(task)
            }
        }
    }

    @PreDestroy
    fun close() {
        actor.close()
        worker.close()
        ioWorker.close()
    }

    /**
     * The state in this internal class is thread safe.
     * See https://kotlinlang.org/docs/shared-mutable-state-and-concurrency.html#actors
     * It is used to manage e.g. fixed task sequences, as well as react to general events.
     */
    private inner class MyActor {

        private val logger = getLogger(javaClass)
        private val groupTaskManager = GroupTaskManager(::onReceive)
        suspend fun onReceive(event: Event) {
            logger.info("[${Thread.currentThread().name}] add task called in Task Manager $event")

            when (event) {
                is RepoChangedEvent -> {
                    ioWorker.addTask(
                        GetGitlabProjectTask(
                            event.repoId,
                            openCodeGitlabConfiguration,
                            ::addEvent,
                            repositoryService,
                        )
                    )
                }

                is GetGitlabProjectDoneEvent -> {

                    ioWorker.addTask(
                        CloneGitTask(
                            event.gitRepository,
                            event.repoId,
                            ::addEvent,
                            config.gitCloneTargetDirectory
                        )
                    )
                }

                is GitCloneDoneEvent -> {

                    val groupId = groupTaskManager.createTaskGroup(event.repoId)

                    val occmdTask = OccmdTask(
                        event.outputDirectory,
                        config.toolResultsTargetDirectory,
                        openCodeGitlabConfiguration,
                        repositoryService,
                        ::addEvent,
                        event.repoId,
                        groupId,
                        toolRunService
                    )

                    val ortAnalyzerTask = OrtAnalyzerTask(
                        event.outputDirectory,
                        config.toolResultsTargetDirectory,
                        ::addEvent,
                        event.repoId,
                        groupId
                    )

                    groupTaskManager.addTaskToGroup(groupId, ortAnalyzerTask.taskID)
                    groupTaskManager.addTaskToGroup(groupId, occmdTask.taskID)


                    worker.addTask(ortAnalyzerTask)
                    worker.addTask(occmdTask)
                }

                is OrtAnalyzerDoneEvent -> {
                    logger.info("Ort analyzer done event received")
                    val ortAdvisorTask = OrtAdvisorTask(
                        event.resultFileOutputDirectoryPath,
                        config.toolResultsTargetDirectory,
                        ::addEvent,
                        event.repoId,
                        event.groupID,
                        dependencyService,
                        event.taskId,
                        toolService,
                        repositoryService
                    )
                    worker.addTask(ortAdvisorTask)
                }

                is GroupTaskDoneEvent -> {
                    logger.info("GroupTaskDoneEvent received")
                    groupTaskManager.taskInGroupFinished(event.groupId, event.taskId)
                }

                is GroupTasksDoneEvent -> {
                    logger.info("Adding repository details task")
                    worker.addTask(
                        GetRepositoryDetailsTask(
                            event.repoId,
                            openCodeGitlabConfiguration,
                            repositoryService,
                            ::addEvent
                        )
                    )
                }

                is GetRepositoryDetailsDoneEvent -> {
                    worker.addTask(
                        MetricsTask(
                            event.repoId,
                            repositoryService,
                            kpiService,
                            ::addEvent,
                            dependencyService,
                            toolRunService
                        )
                    )
                }

                is RecalculateAllKpisEvent -> {
                    repositoryService.getAllRepositories().forEach { repo ->
                        worker.addTask(
                            MetricsTask(
                                repo.id!!,
                                repositoryService,
                                kpiService,
                                ::addEvent,
                                dependencyService,
                                toolRunService
                            )
                        )
                    }
                }

                is TaskFailedEvent -> {
                    logger.info("task failed event received. ${event.throwable}")
                }

                is GroupTaskFailedEvent -> {
                    logger.info("Group task failed event received. ${event.throwable} ")
                    groupTaskManager.taskInGroupFinished(event.groupId, event.taskId)
                }

                else -> {
                    logger.info("Received event without special handling associated $event")
                }
            }
            logger.info("[${Thread.currentThread().name}] add task finished in Task Manager $event")
        }

    }
}
