package de.fraunhofer.iem.dataprovider.taskManager.tasks.kpiCalculation

import de.fraunhofer.iem.dataprovider.dependency.enumeration.VulnerabilityScoringSystemEnum
import de.fraunhofer.iem.dataprovider.dependency.service.DependencyService
import de.fraunhofer.iem.dataprovider.kpi.dto.KPICreateDto
import de.fraunhofer.iem.dataprovider.kpi.service.KPIService
import de.fraunhofer.iem.dataprovider.kpi.strategy.AggregationKPICalculationStrategy
import de.fraunhofer.iem.dataprovider.kpi.strategy.RatioKPICalculationStrategy
import de.fraunhofer.iem.dataprovider.kpi.strategy.RawValueKPICalculationStrategy
import de.fraunhofer.iem.dataprovider.repository.entity.RepositoryDetailsEntity
import de.fraunhofer.iem.dataprovider.repository.service.RepositoryService
import de.fraunhofer.iem.dataprovider.taskManager.events.Event
import de.fraunhofer.iem.dataprovider.taskManager.tasks.Task
import de.fraunhofer.iem.dataprovider.toolRun.service.ToolRunService
import java.util.*

class MetricsTask(
        val repoId: UUID,
        private val repositoryService: RepositoryService,
        private val kpiService: KPIService,
        override val responseChannel: suspend (task: Event) -> Unit,
        private val dependencyService: DependencyService,
        private val toolRunService: ToolRunService,
) : Task() {
    override suspend fun execute() {
        logger.info("Starting metrics task for repoId $repoId")
        val repository = repositoryService.findRepoByID(repoId)
        val repositoryDetails = repositoryService.getLatestRepositoryDetailsByRepositoryId(repoId)
        if (repositoryDetails != null && repository != null) {
            val rootKPI = generateKPITree(repositoryDetails)
            kpiService.calculateKPIs(rootKPI)
            kpiService.storeAndPurgeOld(repository, rootKPI)
        }
    }

    private fun generateKPITree(repositoryDetailsEntity: RepositoryDetailsEntity): KPICreateDto {
        val rootKPI = KPITreeFactory(repositoryDetailsEntity)

        return rootKPI
    }

    private fun KPITreeFactory(repositoryDetailsEntity: RepositoryDetailsEntity): KPICreateDto {
        // lowest level leaves
        val numberOfCommitsKPI = KPICreateDto(
            "Number of Commits",
            "Total number of commits on the default branch of the repository.",
            false,
            RawValueKPICalculationStrategy(repositoryDetailsEntity.numberOfCommits!!)
        )
        val numberOfSignedCommitsKPI = KPICreateDto(
            "Number of Signed Commits",
            "Total number of signed and verified commits on the default branch of the repository.",
            false,
            RawValueKPICalculationStrategy(repositoryDetailsEntity.numberOfSignedCommits!!)
        )
        val isDefaultBranchProtectedKPI = KPICreateDto(
            "Default Branch Protection",
            "Used to assess compliance with a standard development process. For this purpose, it is examined whether the standard development branch is protected against unintentional changes.",
            false,
            RawValueKPICalculationStrategy(if (repositoryDetailsEntity.isDefaultBranchProtected == true) 100 else 0)
        )

        val signedCommitsRatioKPI =
            KPICreateDto(
                "Commit Signature Ratio",
                "Used to assess compliance with a common and transparent development process. It is desirable that all commits are signed by their authors. Therefore, the ratio of signed commits to all commits is determined to calculate this metric.",
                false,
                RatioKPICalculationStrategy()
            )
        signedCommitsRatioKPI.addChildKPI(numberOfCommitsKPI, 1.0)
        signedCommitsRatioKPI.addChildKPI(numberOfSignedCommitsKPI, 1.0)

        // second level
        // TODO: @Jan-Niclas Please add the dependency kpis to the security KPI
        val securityKPI = KPICreateDto(
            "Security Score",
            "Assesses the security of the software provided. For this purpose, various security-relevant analyzes are carried out, which, among other things, check the external dependencies or the code for vulnerabilities.",
            false,
            AggregationKPICalculationStrategy()
        )


        val highestCvssKPI = KPICreateDto(
            "Maximal Dependency Vulnerability Score",
            "This score is calculated by the following formula: 100 - (max(CVSS score) * 10). Thus, a lower value indicates a more critical vulnerability.",
            false,
            AggregationKPICalculationStrategy()
        )
        val dependencyDtos = dependencyService.getDependenciesForRepository(repoId)
        val dependencies: MutableList<KPICreateDto> = mutableListOf()
        var highestScoreDependency: KPICreateDto? = null
        var highestDependencyScore = 0.0

        dependencyDtos.forEach { dependency ->
            val vulnerabilities: MutableList<KPICreateDto> = mutableListOf()
            var highestVulnerability: KPICreateDto? = null
            var highestVulnerabilityScore = 0.0

            dependency.vulnerabilities.forEach { vulnerability ->
                var highestScore = vulnerability.vulnerabilityScores.first()
                vulnerability.vulnerabilityScores.forEach { score ->
                    if (score.scoringSystem == VulnerabilityScoringSystemEnum.CVSSV3_1) {
                        try {
                            val severity = score.severity.toDouble()

                            if ((highestScore.severity.toDouble()) < severity) {
                                highestScore = score
                            }
                        } catch (_: NumberFormatException) {
                        }
                    }
                }
                val highestScoreValue = highestScore.severity.toDouble()
                val kpiValue = 100 - ((highestScoreValue) * 10).toInt()
                val vulnerabilityKpi = KPICreateDto(
                    vulnerability.cveIdentifier,
                    "CVSS V3 Score $highestScoreValue",
                    false,
                    RawValueKPICalculationStrategy(kpiValue),
                    displayScore = "CVSS Rating: ${highestScore.severity}"
                )
                if (highestVulnerabilityScore < highestScoreValue) {
                    highestVulnerability = vulnerabilityKpi
                    highestVulnerabilityScore = highestScoreValue
                }

                vulnerabilities.add(vulnerabilityKpi)
            }

            if (dependency.vulnerabilities.isNotEmpty() && highestVulnerability != null) {
                val dependencyKpi = KPICreateDto(dependency.name, "", false, AggregationKPICalculationStrategy())
                dependencyKpi.addChildKPI(highestVulnerability!!, 1.0)
                if (highestVulnerabilityScore > highestDependencyScore) {
                    highestDependencyScore = highestVulnerabilityScore
                    highestScoreDependency = dependencyKpi
                }
                vulnerabilities.remove(highestVulnerability)
                vulnerabilities.forEach { dependencyKpi.addChildKPI(it, 0.0) }
                dependencies.add(dependencyKpi)
            }
        }

        if (highestScoreDependency != null) {
            highestCvssKPI.addChildKPI(highestScoreDependency!!, 1.0)
            dependencies.remove(highestScoreDependency)
        }
        dependencies.forEach {
            highestCvssKPI.addChildKPI(it, 0.0)
        }

        securityKPI.addChildKPI(highestCvssKPI, 0.5)
        val tr = toolRunService.getLatestToolRunsForRepo(repoId).filter { it.tool?.name == "OCCMD" }[0]
        val secretScore = if (tr.findings.isNotEmpty()) {
            0
        } else {
            100
        }
        val secretKpi = KPICreateDto(
            "Public Secrets",
            "Score to indicate whether the repository contains secrets like passwords or api keys.",
            false,
            RawValueKPICalculationStrategy(secretScore)
        )

        securityKPI.addChildKPI(secretKpi, 0.5)

        val processComplianceKPI = KPICreateDto(
            "Process Compliance Score",
            "Assesses the development process of the software provided. For this purpose, the development process traceable in the repository is compared with common development standards to enable an assessment.",
            false,
            AggregationKPICalculationStrategy()
        )

        val processTransparencyKPI = KPICreateDto(
            "Process Transparency Score",
            "Assesses the transparency resp. traceability of the development process of the provided software for external parties. For this purpose, various analyzes are performed that assess the availability of information about the software development process within the repository.",
            false,
            AggregationKPICalculationStrategy()
        )
        processComplianceKPI.addChildKPI(isDefaultBranchProtectedKPI, 0.5)
        processComplianceKPI.addChildKPI(signedCommitsRatioKPI, 0.5)
        processTransparencyKPI.addChildKPI(signedCommitsRatioKPI, 1.0)

        val rootKPI = KPICreateDto(
            "Project Score",
            "Assesses the project resp. the provided software in the aspects of maturity (based on quality, security and usability aspects) as well as development process.",
            true,
            AggregationKPICalculationStrategy()
        )
        rootKPI.addChildKPI(processTransparencyKPI, 0.25)
        rootKPI.addChildKPI(processComplianceKPI, 0.25)
        rootKPI.addChildKPI(securityKPI, 0.5)
        return rootKPI
    }
}
