package de.fraunhofer.iem.dataprovider.taskManager.tasks.tools.ort

import de.fraunhofer.iem.dataprovider.dependency.dto.DependencyCreateDto
import de.fraunhofer.iem.dataprovider.dependency.dto.VulnerabilityCreateDto
import de.fraunhofer.iem.dataprovider.dependency.dto.VulnerabilityScoreCreateDto
import de.fraunhofer.iem.dataprovider.dependency.entity.DependencyEntity
import de.fraunhofer.iem.dataprovider.dependency.enumeration.VulnerabilityScoringSystemEnum
import de.fraunhofer.iem.dataprovider.dependency.service.DependencyService
import de.fraunhofer.iem.dataprovider.repository.service.RepositoryService
import de.fraunhofer.iem.dataprovider.taskManager.events.Event
import de.fraunhofer.iem.dataprovider.taskManager.tasks.tools.ToolProcessTask
import de.fraunhofer.iem.dataprovider.taskManager.tasks.tools.ort.dto.OrtResultDto
import de.fraunhofer.iem.dataprovider.taskManager.tasks.tools.ort.json.AdvisorResultJson
import de.fraunhofer.iem.dataprovider.taskManager.tasks.tools.ort.json.OrtJson
import de.fraunhofer.iem.dataprovider.tool.dto.CreateToolDto
import de.fraunhofer.iem.dataprovider.tool.service.ToolService
import kotlinx.serialization.json.*
import org.springframework.core.io.ClassPathResource
import org.springframework.core.io.Resource
import java.io.IOException
import java.nio.file.Path
import java.nio.file.Paths
import java.util.*

class OrtAdvisorTask(
    analyzerResultLocation: String,
    outputDirectoryPath: String,
    override val responseChannel: suspend (event: Event) -> Unit,
    override val repoId: UUID,
    override val groupID: UUID?,
    private val dependencyService: DependencyService,
    override val taskID: UUID,
    private val toolService: ToolService,
    private val repositoryService: RepositoryService
) : ToolProcessTask<List<DependencyCreateDto>>() {

    private val resource: Resource = ClassPathResource("scripts/ort/ort_advisor.sh")
    override val outputDirectory = Paths.get(outputDirectoryPath, "ort-advisor", taskID.toString()).toString()
    override val resultFileOutputPath: Path = Paths.get(outputDirectory, "advisor-result.json")

    override val flags = arrayOf(resource.file.absolutePath, analyzerResultLocation, outputDirectory)

    override suspend fun parseProcessResults(resultPath: Path): List<DependencyCreateDto> {
        val dependencies = mutableListOf<DependencyCreateDto>()
        val ortJson = getOrtResultsFromFile(resultPath)

        ortJson.packages.forEach { p ->
            if (p.nameAndVersion != null) {
                dependencies.add(DependencyCreateDto(name = p.nameAndVersion))
            }
        }

        ortJson.advisorResults.forEach { advisorResult ->

            var dependency = dependencies.find { it.name == advisorResult.identifier }

            if (dependency == null) {
                dependency = DependencyCreateDto(advisorResult.identifier)
                dependencies.add(dependency)
            }
            advisorResult.vulnerabilities.forEach { vulnerability ->
                if (vulnerability.id != null) {

                    val vulnerabilityScoreDtos =
                        vulnerability.references.mapNotNull { ref ->
                            if (ref.severity != null && ref.scoringSystem != null) {
                                val scoringSystem = VulnerabilityScoringSystemEnum.fromString(ref.scoringSystem)
                                scoringSystem?.let {
                                    return@mapNotNull VulnerabilityScoreCreateDto(severity = ref.severity, it)
                                }
                            }
                            return@mapNotNull null
                        }
                    val vulnerabilityDto =
                        VulnerabilityCreateDto(vulnerability.id, vulnerabilityScoreDtos.toMutableList())
                    dependency.vulnerabilities.add(vulnerabilityDto)
                }
            }
        }

        return dependencies
    }

    // TODO: Requires a database purge / association with the tool run
    override suspend fun storeResultsInDb(result: List<DependencyCreateDto>) {
        val dependencies = mutableListOf<DependencyEntity>()
        result.forEach { dependency ->
            val dependencyEntity = dependency.toDbObject()
            dependencyService.save(dependencyEntity)
            dependencies.add(dependencyEntity)
        }
        val tool = toolService.findOrCreateTool(CreateToolDto("ORT", "ORT", "docker-snapshot"))
        val repo = repositoryService.findRepoByID(repoId)
        if (repo != null) {
            repo.toolEntities.add(tool)
            repo.dependencyEntities.addAll(dependencies)
            repositoryService.save(repo)
        }

    }

}

fun getOrtResultsFromFile(resultPath: Path): OrtResultDto {
    val resFile = resultPath.toFile()

    if (resFile.exists()) {
        val results = mutableListOf<AdvisorResultJson>()
        val resString = resFile.readText()
        val json = Json { ignoreUnknownKeys = true }

        val rawJson = json.decodeFromString<OrtJson>(resString)
        rawJson.advisor?.results?.advisorResults?.forEach { (key, value) ->
            if (value is JsonArray) {
                value.forEach { res ->
                    if (res is JsonObject) {

                        val flat = buildJsonObject {
                            res.entries.forEach { (eKey, eValue) ->
                                put(eKey, eValue)
                            }
                            put("identifier", key)
                        }

                        val advisorResult = json.decodeFromJsonElement<AdvisorResultJson>(flat)
                        results.add(advisorResult)
                    }
                }
            }
        }
        val packages = rawJson.analyzer?.result?.packages?.toList() ?: listOf()
        return OrtResultDto(advisorResults = results, packages = packages)
    }

    throw IOException("File not found / Path is no file.")
}
